/************
Virtu@L Dream, DreamCast emulator for PC
Copyright (C) 2003 eBrain  H.H.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
************/

#include <GL/glut.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#define SH4DSM

#ifndef _linux
#if defined (__linux__)
 #define _linux
#else
 #define WINDOWS
#endif
#endif

#if defined _linux
 #include <pthread.h>
#elif  defined WINDOWS
 #include <winbase.h>
#endif


#ifndef BYTE
 #define BYTE unsigned char
#endif
#ifndef WORD
 #define WORD unsigned short
#endif
#ifndef LONG
 #define LONG unsigned long
#endif
#ifndef LONG64
#if defined _linux
 #define LONG64 unsigned long long
#elif defined WINDOWS
 #define LONG64 unsigned _int64
#endif
#endif

//CPU
//registers
extern unsigned long R[16];
extern unsigned long R_BANK[8];
extern unsigned long SR,FPSCR,GBR,VBR,SSR,SPC,DBR,SGR;
extern unsigned long MACH,MACL,PR;
extern unsigned long PC;
extern float FPUL;
//MM registers
extern BYTE CCN[0x40];
extern BYTE UBC[0x22];
extern BYTE BSC1[0x50];
extern BYTE BSC2[0x10000];
extern BYTE BSC3[0x10000];
extern BYTE DMAC[0x44];
extern BYTE CPG[0x11];
extern BYTE RTC[0x3D];
extern BYTE INTC[0xE];
extern BYTE TMU[0x30];
extern BYTE SCI[0x1D];
extern BYTE SCIF[0x26];
extern BYTE HUDI[0xC];
extern BYTE *mmregtab[0xf1];

extern unsigned long *PTEH;
extern unsigned long *PTEL;
extern unsigned long *TTB;
extern unsigned long *TEA;
extern unsigned long *MMUCR;
extern unsigned long *TRA;
extern unsigned long *EXPEVT;
extern unsigned long *INTEVT;
extern unsigned long *PTEA;
extern unsigned long *CCR;

struct MMUCR0
{
unsigned long     AT:1;
unsigned long dummy0:1;
unsigned long     TI:1;
unsigned long dummy1:5;
unsigned long     SV:1;
unsigned long dummy2:1;
unsigned long     URC:6;
unsigned long dummy3:2;
unsigned long     URB:6;
unsigned long dummy4:2;
unsigned long     LRUI:6;
};

#define MMUCR_URC ((*(struct MMUCR0*)(MMUCR)).URC)
//SR structure definitions
struct SR0 {
unsigned long     T0:1;
unsigned long     S0:1;
unsigned long dummy0:2;
unsigned long     I0:4;
unsigned long     Q0:1;
unsigned long     M0:1;
unsigned long dummy1:5;
unsigned long     FD:1;
unsigned long dummy2:12;
unsigned long     BL:1;
unsigned long     RB:1;
unsigned long     MD:1;
unsigned long dummy3:1;
};
#define MD ((*(struct SR0 *)(&SR)).MD)
#define RB ((*(struct SR0 *)(&SR)).RB)
#define BL ((*(struct SR0 *)(&SR)).BL)
#define FD ((*(struct SR0 *)(&SR)).FD)
#define M ((*(struct SR0 *)(&SR)).M0)
#define Q ((*(struct SR0 *)(&SR)).Q0)
#define IM ((*(struct SR0 *)(&SR)).I0)
#define S ((*(struct SR0 *)(&SR)).S0)
#define T ((*(struct SR0 *)(&SR)).T0)

struct PTEH0
{
    unsigned long ASID:8;
    unsigned long dummy0:2;
    unsigned long VPN:22;
};

#define PTEH_VPN ((*(struct PTEH0 *)(PTEH)).VPN)

#define CCR_OIX (*CCR>>7)&1

struct sTLB
{
 unsigned char ASID;
 int VPN;
 int PPN;
 unsigned char SZ,SH,PR,WT,C,D,V,SA,TC;
};

extern struct sTLB TLB[64];



void Delay_Slot(unsigned  long Addr);
void Error( char *er ); //Error display function
BYTE Read_Byte(unsigned  long Addr);
WORD Read_Word(unsigned  long Addr);
LONG Read_Long(unsigned  long Addr);
void Write_Byte(unsigned  long Addr, BYTE Data);
void Write_Word(unsigned  long Addr, WORD Data);
void Write_Long(unsigned  long Addr, LONG Data);
void Sleep_standby();
int  allocate_operand_cache_block(int n);
void check_single_exception(float *dst,float result);
void check_double_exception(double *dst,double result);
int  check_product_invalid(int m,int n);
int  check_product_infinity(int m,int n);
int  check_positive_infinity(int m,int n);
int  check_negative_infinity(int m,int n);
void clear_cause();
int  data_type_of(int n);
void dz(int n,int sign);
void fcmp_invalid();
void fcnvds(int m);
void fcnvsd(int n);
void fipr(int m,int n);
void fpu_exception_trap();
int  fpul_type();
void ftrc_invalid(int sign);
int  ftrc_single_type_of(int m);
int  ftrc_double_type_of(int m);
void inexact();
void inf(int n,int sign);
void invalid(int n);
int  invalidate_operand_cache_block(int n);
int  is_dirty_block(int n);
int  is_write_back_memory(int n);
int  look_up_in_operand_cache(int n);
void normal_faddsub(int m,int n,int type);
void normal_fcnvds(int m);
void normal_fdiv(int m,int n);
void normal_fmul(int m,int n);
void normal_fmac(int m,int n);
void normal_fsqrt(int n);
void set_E();
void set_V();
void set_Z();
void set_O();
void set_U();
void set_I();
void qnan(int n);
void register_copy(int m,int n);
int  sign_of(int n);
void undefined_operation();
int  write_back(int n);
void zero(int n,int sign);

//Emulation Functions
void loadip_1stread(char *);
void mapdsmall();
void Power_on_reset();
void makepvrinfo(unsigned long Addr, LONG Data);
void mem_map_reg();
void create_on_chip_ram();
void create_ram();
void *glstart(void *);


//FPU
#define PZERO       0
#define NZERO       1
#define DENORM      2
#define NORM        3
#define PINF        4
#define NINF        5
#define qNaN        6
#define sNaN        7
#define EQ          0
#define GT          1
#define LT          2
#define UO          3
#define INVALID     4
#define FADD        0
#define FSUB        1


#define CAUSE       0x0003f000  /* FPSCR(bit17-12) */
#define SET_E       0x00020000  /* FPSCR(bit17) */
#define SET_V       0x00010040  /* FPSCR(bit16,6) */
#define SET_Z       0x00008020  /* FPSCR(bit15,5) */
#define SET_O       0x00004010  /* FPSCR(bit14,4) */
#define SET_U       0x00002008  /* FPSCR(bit13,3) */
#define SET_I       0x00001004  /* FPSCR(bit12,2) */
#define ENABLE_VOUI 0x00000b80  /* FPSCR(bit11,9-7) */
#define ENABLE_V    0x00000800  /* FPSCR(bit11) */
#define ENABLE_Z    0x00000400  /* FPSCR(bit10) */
#define ENABLE_OUI  0x00000380  /* FPSCR(bit9-7) */
#define ENABLE_I    0x00000080  /* FPSCR(bit7) */
#define FLAG        0x0000007C  /* FPSCR(bit6-2) */

extern union {
        int  l[2][16];
        float f[2][16];
        double d[2][8];
}frf;


#define FPSCR_FR    FPSCR>>21&1
#define FPSCR_SZ    FPSCR>>20&1
#define FPSCR_PR    FPSCR>>19&1
#define FPSCR_DN    FPSCR>>18&1
#define FPSCR_I     FPSCR>>12&1
#define FPSCR_RM    FPSCR&1
#define FR          frf.f[ FPSCR_FR]
#define FR_HEX      frf.l[ FPSCR_FR]
#define DR          frf.d[ FPSCR_FR]
#define DR_HEX      frf.d[ FPSCR_FR]
#define XF          frf.f[~FPSCR_FR]
#define XF_HEX      frf.l[~FPSCR_FR]
#define XD          frf.d[~FPSCR_FR]



//MEM
#define EntryPoint8C 0x8C000000
#define EntryPointAC 0xAC000000
#define EntryPointA05 0xA0500000
#define EntryPointA07 0xA0700000
#define EntryPointA08 0xA0800000
#define EntryPointA4 0xA4000000
#define EntryPointA5 0xA5000000
#define IPOFFSET 0x8000
#define BOOTOFFSET 0x10000
#define WSIZEX 640
#define WSIZEY 480
#define FrameBufSIZE 1228800 //640*480*4
#define RAMONCHIPSIZE 0x1000


extern BYTE RamAlloc; //use to know if allow or not mem
//2*16*4K
extern BYTE *RamArea1,*RamArea2;
//4*16*16=1024
extern BYTE *IndexRamArea[0x10];

//PAGE 50
extern BYTE *DREAMCAST_PRAM; //16 M
extern BYTE *DREAMCAST_VRAM; //8 M
extern BYTE *DREAMCAST_RAM_A05; //1 M
extern BYTE *DREAMCAST_RAM_A07; //1 M
extern BYTE *DREAMCAST_RAM_A08; //1 M
extern GLubyte *FrameStart;
extern GLubyte *FrameBufAddr;



//EX_INT
extern char *ROMNAME;


struct tagfpar
{
   void (*fonc)();
   int par1;
   int par2;
   int par3;
};
extern struct tagfpar tab_op[0x10000];


//OPCODES
  void ADD(long m, long n); /* ADD Rm,Rn */
  void ADDI(long i, long n);  /* ADD #imm,Rn */
  void ADDC(long m, long n);    /* ADDC Rm,Rn */
  void ADDV(long m, long n);    /* ADDV Rm,Rn */
  void AND(long m, long n);    /* AND Rm,Rn */
  void ANDI(long i);    /* AND #imm,R0 */
  void ANDM(long i);    /* AND.B #imm,@(R0,GBR); */
  void BF(int d);    /* BF disp */
  void BFS(int d);    /* BFS disp */
  void BRA(int d);    /* BRA disp */
  void BRAF(int n);   /* BRAF Rn */
  void BSR(int d);   /* BSR disp */
  void BSRF(int n);     /* BSRF Rn */
  void BT(int d);    /* BT disp */
  void BTS(int d);    /* BTS disp */
  void CLRMAC();    /* CLRMAC */
  void CLRS();    /* CLRS */
  void CLRT();    /* CLRT */
  void CMPEQ(long m, long n);    /* CMP_EQ Rm,Rn */
  void CMPGE(long m, long n);    /* CMP_GE Rm,Rn */
  void CMPGT(long m, long n);    /* CMP_GT Rm,Rn */
  void CMPHI(long m, long n);    /* CMP_HI Rm,Rn */
  void CMPHS(long m, long n);    /* CMP_HS Rm,Rn */
  void CMPPL(long n);              /* CMP_PL Rn */
  void CMPPZ(long n);    /* CMP_PZ Rn */
  void CMPSTR(long m, long n);    /* CMP_STR Rm,Rn */
  void CMPIM(long i);    /* CMP_EQ #imm,R0 */
  void DIV0S(long m, long n);    /* DIV0S Rm,Rn */
  void DIV0U();        /* DIV0U */
  void DIV1(long m, long n);    /* DIV1 Rm,Rn */
  void DMULS(long m, long n);  /* DMULS.L Rm,Rn */
  void DMULU(long m, long n);  /* DMULU.L Rm,Rn */
  void DT(long n);/* DT Rn */
  void EXTSB(long m, long n);            /* EXTS.B Rm,Rn */
  void EXTSW(long m, long n);            /* EXTS.W Rm,Rn */
  void EXTUB(long m, long n);             /* EXTU.B Rm,Rn */
  void EXTUW(long m, long n);             /* EXTU.W Rm,Rn */
  void JMP(int n);/* JMP @Rn */
  void JSR(int n);/* JSR @Rn */
  void LDCSR(int m);      /* LDC Rm,SR : Privileged */
  void LDCGBR(int m);     /* LDC Rm,GBR */
  void LDCVBR(int m);      /* LDC Rm,VBR : Privileged */
  void LDCSSR(int m);      /* LDC Rm,SSR : Privileged */
  void LDCSPC(int m);      /* LDC Rm,SPC : Privileged */
  void LDCDBR(int m);      /* LDC Rm,DBR : Privileged */
  void LDCRn_BANK(int m, int n);  /* LDC Rm,Rn_BANK : Privileged */
  void LDCMSR(int m);     /* LDC.L @Rm+,SR : Privileged */
  void LDCMGBR(int m);     /* LDC.L @Rm+,GBR */
  void LDCMVBR(int m);     /* LDC.L @Rm+,VBR : Privileged */
  void LDCMSSR(int m);     /* LDC.L @Rm+,SSR : Privileged */
  void LDCMSPC(int m);     /* LDC.L @Rm+,SPC : Privileged */
  void LDCMDBR(int m);     /* LDC.L @Rm+,DBR : Privileged */
  void LDCMRn_BANK(int m,int n); /* LDC.L @Rm+,Rn_BANK : Privileged ;  n=07 */
  void LDSFPUL(int m);        /* LDS Rm,FPUL  */
  void LDSMFPUL(int m);        /* LDS.L @Rm+,FPUL  */
  void LDSFPSCR(int  m);        /* LDS Rm,FPSCR  */
  void LDSMFPSCR(int  m);        /* LDS.L @Rm+,FPSCR  */
  void LDSMACH(int m);     /* LDS Rm,MACH */
  void LDSMACL(int m);     /* LDS Rm,MACL */
  void LDSPR(int m);      /* LDS Rm,PR */
  void LDSMMACH(int m);     /* LDS.L @Rm+,MACH */
  void LDSMMACL(int m);     /* LDS.L @Rm+,MACL */
  void LDSMPR(int m);     /* LDS.L @Rm+,PR */
  void LDTLB(); /* LDTLB */
  void MAC_L(long m, long n);   /* MAC.L @Rm+,@Rn+ */
  void MACW(long m, long n);   /* MAC.W @Rm+,@Rn+ */
  void MOV(long m, long n);   /* MOV Rm,Rn */
  void MOVBS(long m, long n); /* MOV.B Rm,@Rn */
  void MOVWS(long m, long n); /* MOV.W Rm,@Rn */
  void MOVLS(long m, long n); /* MOV.L Rm,@Rn */
  void MOVBL(long m, long n); /* MOV.B @Rm,Rn */
  void MOVWL(long m, long n); /* MOV.W @Rm,Rn */
  void MOVLL(long m, long n); /* MOV.L @Rm,Rn */
  void MOVBM(long m, long n); /* MOV.B Rm,@-Rn */
  void MOVWM(long m, long n); /* MOV.W Rm,@-Rn */
  void MOVLM(long m, long n);   /* MOV.L Rm,@-Rn */
  void MOVBP(long m, long n); /* MOV.B @Rm+,Rn */
  void MOVWP(long m, long n);    /* MOV.W @Rm+,Rn */
  void MOVLP(long m, long n);   /* MOV.L @Rm+,Rn */
  void MOVBS0(long m, long n); /* MOV.B Rm,@(R0,Rn); */
  void MOVWS0(long m, long n); /* MOV.W Rm,@(R0,Rn); */
  void MOVLS0(long m, long n); /* MOV.L Rm,@(R0,Rn); */
  void MOVBL0(long m, long n); /* MOV.B @(R0,Rm);,Rn */
  void MOVWL0(long m, long n);   /* MOV.W @(R0,Rm);,Rn */
  void MOVLL0(long m, long n);   /* MOV.L @(R0,Rm);,Rn */
  void MOVI(int i, int n);  /* MOV #imm,Rn */
  void MOVWI(int d, int n); /* MOV.W @(disp,PC);,Rn */
  void MOVLI(int d, int n);/* MOV.L @(disp,PC);,Rn */
  void MOVBLG(int d); /* MOV.B @(disp,GBR);,R0 */
  void MOVWLG(int d); /* MOV.W @(disp,GBR);,R0 */
  void MOVLLG(int d); /* MOV.L @(disp,GBR);,R0 */
  void MOVBSG(int d); /* MOV.B R0,@(disp,GBR); */
  void MOVWSG(int d); /* MOV.W R0,@(disp,GBR); */
  void MOVLSG(int d); /* MOV.L R0,@(disp,GBR); */
  void MOVBS4(long d, long n);   /* MOV.B R0,@(disp,Rn); */
  void MOVWS4(long d, long n);  /* MOV.W R0,@(disp,Rn); */
  void MOVLS4(long m, long d, long n); /* MOV.L Rm,@(disp,Rn); */
  void MOVBL4(long m, long d);  /* MOV.B @(disp,Rm);,R0 */
  void MOVWL4(long m, long d); /* MOV.W @(disp,Rm);,R0 */
  void MOVLL4(long m, long d, long n); /* MOV.L @(disp,Rm);,Rn */
  void MOVA(int d);               /* MOVA @(disp,PC);,R0 */
  void MOVCAL(int n);      //MOVCA.L  R0,@Rn
  void MOVT(long n);               /* MOVT Rn */
  void MULL(long m, long n); /* MUL.L Rm,Rn */
  void MULS(long m, long n); /* MULS Rm,Rn */
  void MULU(long m, long n); /* MULU Rm,Rn  */
  void NEG(long m, long n); /* NEG Rm,Rn */
  void NEGC(long m, long n); /* NEGC Rm,Rn */
  void NOP(); /* NOP */
  void NOT(long m, long n); /* NOT Rm,Rn */
  void OCBI(int n);        /* OCBI @Rn */
  void OCBP(int n);        /* OCBP @Rn */
  void OCBWB(int n);       /* OCBWB @Rn */
  void OR(long m, long n); /* OR Rm,Rn */
  void ORI(long i);  /* OR #imm,R0 */
  void ORM(long i); /* OR.B #imm,@(R0,GBR); */
  void PREF(int n); /* PREF */
  void ROTCL(long n); /* ROTCL Rn */
  void ROTCR(long n); /* ROTCR Rn */
  void ROTL(long n);  /* ROTL Rn */
  void ROTR(long n); /* ROTR Rn */
  void RTE(); /* RTE */
  void RTS(); /* RTS */
  void SETS(); /* SETS */
  void SETT();  /* SETT */
  void SHAD(int m,int n); /*SHAD Rm,Rn */
  void SHAL(long n); /* SHAL Rn (Same as SHLL); */
  void SHAR(long n); /* SHAR Rn */
  void SHLD(int m,int n);/*SHLD Rm,Rn */
  void SHLL(long n); /* SHLL Rn  (Same as SHAL); */
  void SHLL2(long n); /* SHLL2 Rn */
  void SHLL8(long n); /* SHLL8 Rn */
  void SHLL16(long n); /* SHLL16 Rn */
  void SHLR(long n);   /* SHLR Rn */
  void SHLR2(long n);          /* SHLR2 Rn */
  void SHLR8(long n);          /* SHLR8 Rn */
  void SHLR16(long n);         /* SHLR16 Rn */
  void SLEEP();    /* SLEEP */
  void STCSR(int n);      /* STC SR,Rn : Privileged */
  void STCGBR(int n);     /* STC GBR,Rn */
  void STCVBR(int n);     /* STC VBR,Rn : Privileged */
  void STCSSR(int n);     /* STC SSR,Rn : Privileged */
  void STCSPC(int n);     /* STC SPC,Rn : Privileged */
  void STCSGR(int n);     /* STC SGR,Rn : Privileged */
  void STCDBR(int n);     /* STC DBR,Rn : Privileged */
  void STCRm_BANK(int n, int m);     /* STC Rm_BANK,Rn : Privileged ; m=07 */
  void STCMSR(int n);     /* STC.L SR,@-Rn : Privileged */
  void STCMGBR(int n);     /* STC.L GBR,@Rn */
  void STCMVBR(int n);     /* STC.L VBR,@-Rn : Privileged */
  void STCMSSR(int n);     /* STC.L SSR,@-Rn : Privileged */
  void STCMSPC(int n);     /* STC.L SPC,@-Rn : Privileged */
  void STCMSGR(int n);     /* STC.L SGR,@-Rn : Privileged */
  void STCMDBR(int n);     /* STC.L DBR,@-Rn : Privileged */
  void STCMRm_BANK(int n, int m);     /* STC.L Rm_BANK,@-Rn : Privileged ; m=07 */
  void STSMACH(int n);     /* STS MACH,Rn */
  void STSMACL(int n);     /* STS MACL,Rn */
  void STSPR(int n);     /* STS PR,Rn */
  void STSMMACH(int n);     /* STS.L MACH,@-Rn */
  void STSMMACL(int n);     /* STS.L MACL,@-Rn */
  void STSMPR(int n);     /* STS.L PR,@-Rn */
  void STS(int n);           /* STS FPUL,Rn  */
  void STS_SAVE(int n);     /* STS.L FPUL,@-Rn  */
  void STS_FPSCR(int n);          /* STS FPSCR,Rn  */
  void STS_RESTORE(int  n);  /* STS.L FPSCR,@-Rn  */
  void SUB(long m, long n); /* SUB Rm,Rn */
  void SUBC(long m, long n); /* SUBC Rm,Rn */
  void SUBV(long m, long n); /* SUBV Rm,Rn */
  void SWAPB(long m, long n);      /* SWAP.B Rm,Rn */
  void SWAPW(long m, long n);      /* SWAP.W Rm,Rn */
  void TAS(int n); /* TAS.B @Rn */
  void TRAPA(int i); /* TRAPA #imm */
  void TST(long m, long n); /* TST Rm,Rn */
  void TSTI(long i); /* TST #imm,R0 */
  void TSTM(long i); /* TST.B #imm,@(R0,GBR); */
  void XOR(long m, long n); /* XOR Rm,Rn */
  void XORI(long i);  /* XOR #imm,R0 */
  void XORM(long i);  /* XOR.B #imm,@(R0,GBR); */
  void XTRCT(long m, long n);     /* XTRCT Rm,Rn */
  void FABS(int n); /* FABS FRn */
  void FLADD(int m,int n); /*FADD FRm, FRn */
  void FCMP_EQ(int m,int n); /* FCMP/EQ  FRm,FRn */
  void FCMP_GT(int m,int n); /* FCMP/GT  FRm,FRn */
  void FCNVDS(int m); /* FCNVDS DRm,FPUL */
  void FCNVSD(int n); /* FCNVSD FPUL, DRn */
  void FDIV(int m,int n);     /* FDIV FRm,FRn */
  void FIPR(int m,int n);   /* FIPR FVm,FVn */
  void FLDI0(int n); /* FLDI0 FRn */
  void FLDI1(int n); /* FLDI1 FRn */
  void FLDS(int m);  /*FLDS FRm,FPUL*/
  void FPUFLOAT(int n); /*FLOAT FPUL,FRn*/
  void FMAC(int m,int n); /*FMAC FR0,FRm,FRn*/
  void FMOV(int m,int n);                 /* FMOV FRm,FRn */
  void FMOV_DR(int m,int n);           /* FMOV DRm,DRn */
  void FMOV_STORE(int m,int n);       /* FMOV.S FRm,@Rn */
  void FMOV_STORE_DR(int m,int n);   /* FMOV DRm,@Rn */
  void FMOV_LOAD(int m,int n);       /* FMOV.S @Rm,FRn */
  void FMOV_LOAD_DR(int m,int n);    /* FMOV @Rm,DRn */
  void FMOV_RESTORE(int m,int n);     /* FMOV.S @Rm+,FRn */
  void FMOV_RESTORE_DR(int m,int n); /* FMOV @Rm+,DRn */
  void FMOV_SAVE(int m,int n);        /* FMOV.S FRm,@Rn */
  void FMOV_SAVE_DR(int m,int n);    /* FMOV DRm,@Rn */
  void FMOV_INDEX_LOAD(int m,int n);  /* FMOV.S @(R0,Rm);,FRn */
  void FMOV_INDEX_LOAD_DR(int m,int n); /*FMOV @(R0,Rm);,DRn */
  void FMOV_INDEX_STORE(int m,int n);  /*FMOV.S FRm,@(R0,Rn);*/
  void FMOV_INDEX_STORE_DR(int m,int n);/*FMOV DRm,@(R0,Rn);*/
  void FMOV_STORE_XD(int m,int n);      /* FMOV XDm,@Rn */
  void FMOV_LOAD_XD(int m,int n);      /* FMOV @Rm,XDn */
  void FMOV_RESTORE_XD(int m,int n);   /* FMOV @Rm+,DBn */
  void FMOV_SAVE_XD(int m,int n);      /* FMOV XDm,@Rn */
  void FMOV_INDEX_LOAD_XD(int m,int n);           /* FMOV @(R0,Rm);,XDn */
  void FMOV_INDEX_STORE_XD(int m,int n);          /* FMOV XDm,@(R0,Rn); */
  void FMOV_XDXD(int m,int n);             /* FMOV XDm,XDn */
  void FMOV_XDDR(int m,int n); /* FMOV XDm,DRn */
  void FMOV_DRXD(int m,int n); /* FMOV DRm,XDn */
  void FMUL(int m,int n); /*FMUL FRm,FRn*/
  void FNEG (int n); /*FNEG FRn*/
  void FSCHG();   /* FSCHG */
  void FRCHG();   /* FRCHG */
  void FSQRT(int n); /*FSQRT FRn*/
  void FSTS(int n); /*FSTS FPUL,FRn*/
  void FLSUB (int m,int n); /*FSUB FRm,FRn */
  void FTRC(int m); /*FTRC FRm,FPUL*/
  void FTRV (int n);     /* FTRV FVn */


//DSM OPCODES
  void DSMADD(long m, long n); /* ADD Rm,Rn */
  void DSMADDI(long i, long n);  /* ADD #imm,Rn */
  void DSMADDC(long m, long n);    /* ADDC Rm,Rn */
  void DSMADDV(long m, long n);    /* ADDV Rm,Rn */
  void DSMAND(long m, long n);    /* AND Rm,Rn */
  void DSMANDI(long i);    /* AND #imm,R0 */
  void DSMANDM(long i);    /* AND.B #imm,@(R0,GBR); */
  void DSMBF(int d);    /* BF disp */
  void DSMBFS(int d);    /* BFS disp */
  void DSMBRA(int d);    /* BRA disp */
  void DSMBRAF(int n);   /* BRAF Rn */
  void DSMBSR(int d);   /* BSR disp */
  void DSMBSRF(int n);     /* BSRF Rn */
  void DSMBT(int d);    /* BT disp */
  void DSMBTS(int d);    /* BTS disp */
  void DSMCLRMAC();    /* CLRMAC */
  void DSMCLRS();    /* CLRS */
  void DSMCLRT();    /* CLRT */
  void DSMCMPEQ(long m, long n);    /* CMP_EQ Rm,Rn */
  void DSMCMPGE(long m, long n);    /* CMP_GE Rm,Rn */
  void DSMCMPGT(long m, long n);    /* CMP_GT Rm,Rn */
  void DSMCMPHI(long m, long n);    /* CMP_HI Rm,Rn */
  void DSMCMPHS(long m, long n);    /* CMP_HS Rm,Rn */
  void DSMCMPPL(long n);              /* CMP_PL Rn */
  void DSMCMPPZ(long n);    /* CMP_PZ Rn */
  void DSMCMPSTR(long m, long n);    /* CMP_STR Rm,Rn */
  void DSMCMPIM(long i);    /* CMP_EQ #imm,R0 */
  void DSMDIV0S(long m, long n);    /* DIV0S Rm,Rn */
  void DSMDIV0U();        /* DIV0U */
  void DSMDIV1(long m, long n);    /* DIV1 Rm,Rn */
  void DSMDMULS(long m, long n);  /* DMULS.L Rm,Rn */
  void DSMDMULU(long m, long n);  /* DMULU.L Rm,Rn */
  void DSMDT(long n);/* DT Rn */
  void DSMEXTSB(long m, long n);            /* EXTS.B Rm,Rn */
  void DSMEXTSW(long m, long n);            /* EXTS.W Rm,Rn */
  void DSMEXTUB(long m, long n);             /* EXTU.B Rm,Rn */
  void DSMEXTUW(long m, long n);             /* EXTU.W Rm,Rn */
  void DSMJMP(int n);/* JMP @Rn */
  void DSMJSR(int n);/* JSR @Rn */
  void DSMLDCSR(int m);      /* LDC Rm,SR : Privileged */
  void DSMLDCGBR(int m);     /* LDC Rm,GBR */
  void DSMLDCVBR(int m);      /* LDC Rm,VBR : Privileged */
  void DSMLDCSSR(int m);      /* LDC Rm,SSR : Privileged */
  void DSMLDCSPC(int m);      /* LDC Rm,SPC : Privileged */
  void DSMLDCDBR(int m);      /* LDC Rm,DBR : Privileged */
  void DSMLDCRn_BANK(int m, int n);  /* LDC Rm,Rn_BANK : Privileged */
  void DSMLDCMSR(int m);     /* LDC.L @Rm+,SR : Privileged */
  void DSMLDCMGBR(int m);     /* LDC.L @Rm+,GBR */
  void DSMLDCMVBR(int m);     /* LDC.L @Rm+,VBR : Privileged */
  void DSMLDCMSSR(int m);     /* LDC.L @Rm+,SSR : Privileged */
  void DSMLDCMSPC(int m);     /* LDC.L @Rm+,SPC : Privileged */
  void DSMLDCMDBR(int m);     /* LDC.L @Rm+,DBR : Privileged */
  void DSMLDCMRn_BANK(int m, int n); /* LDC.L @Rm+,Rn_BANK : Privileged ;  n=07 */
  void DSMLDSFPUL(int m);        /* LDS Rm,FPUL  */
  void DSMLDSMFPUL(int m);        /* LDS.L @Rm+,FPUL  */
  void DSMLDSFPSCR(int  m);        /* LDS Rm,FPSCR  */
  void DSMLDSMFPSCR(int  m);        /* LDS.L @Rm+,FPSCR  */
  void DSMLDSMACH(int m);     /* LDS Rm,MACH */
  void DSMLDSMACL(int m);     /* LDS Rm,MACL */
  void DSMLDSPR(int m);      /* LDS Rm,PR */
  void DSMLDSMMACH(int m);     /* LDS.L @Rm+,MACH */
  void DSMLDSMMACL(int m);     /* LDS.L @Rm+,MACL */
  void DSMLDSMPR(int m);     /* LDS.L @Rm+,PR */
  void DSMLDTLB();
  void DSMMAC_L(long m, long n);   /* MAC.L @Rm+,@Rn+ */
  void DSMMACW(long m, long n);   /* MAC.W @Rm+,@Rn+ */
  void DSMMOV(long m, long n);   /* MOV Rm,Rn */
  void DSMMOVBS(long m, long n); /* MOV.B Rm,@Rn */
  void DSMMOVWS(long m, long n); /* MOV.W Rm,@Rn */
  void DSMMOVLS(long m, long n); /* MOV.L Rm,@Rn */
  void DSMMOVBL(long m, long n); /* MOV.B @Rm,Rn */
  void DSMMOVWL(long m, long n); /* MOV.W @Rm,Rn */
  void DSMMOVLL(long m, long n); /* MOV.L @Rm,Rn */
  void DSMMOVBM(long m, long n); /* MOV.B Rm,@-Rn */
  void DSMMOVWM(long m, long n); /* MOV.W Rm,@-Rn */
  void DSMMOVLM(long m, long n);   /* MOV.L Rm,@-Rn */
  void DSMMOVBP(long m, long n); /* MOV.B @Rm+,Rn */
  void DSMMOVWP(long m, long n);    /* MOV.W @Rm+,Rn */
  void DSMMOVLP(long m, long n);   /* MOV.L @Rm+,Rn */
  void DSMMOVBS0(long m, long n); /* MOV.B Rm,@(R0,Rn); */
  void DSMMOVWS0(long m, long n); /* MOV.W Rm,@(R0,Rn); */
  void DSMMOVLS0(long m, long n); /* MOV.L Rm,@(R0,Rn); */
  void DSMMOVBL0(long m, long n); /* MOV.B @(R0,Rm);,Rn */
  void DSMMOVWL0(long m, long n);   /* MOV.W @(R0,Rm);,Rn */
  void DSMMOVLL0(long m, long n);   /* MOV.L @(R0,Rm);,Rn */
  void DSMMOVI(int i, int n);  /* MOV #imm,Rn */
  void DSMMOVWI(int d, int n); /* MOV.W @(disp,PC);,Rn */
  void DSMMOVLI(int d, int n);/* MOV.L @(disp,PC);,Rn */
  void DSMMOVBLG(int d); /* MOV.B @(disp,GBR);,R0 */
  void DSMMOVWLG(int d); /* MOV.W @(disp,GBR);,R0 */
  void DSMMOVLLG(int d); /* MOV.L @(disp,GBR);,R0 */
  void DSMMOVBSG(int d); /* MOV.B R0,@(disp,GBR); */
  void DSMMOVWSG(int d); /* MOV.W R0,@(disp,GBR); */
  void DSMMOVLSG(int d); /* MOV.L R0,@(disp,GBR); */
  void DSMMOVBS4(long d, long n);   /* MOV.B R0,@(disp,Rn); */
  void DSMMOVWS4(long d, long n);  /* MOV.W R0,@(disp,Rn); */
  void DSMMOVLS4(long m, long d, long n); /* MOV.L Rm,@(disp,Rn); */
  void DSMMOVBL4(long m, long d);  /* MOV.B @(disp,Rm);,R0 */
  void DSMMOVWL4(long m, long d); /* MOV.W @(disp,Rm);,R0 */
  void DSMMOVLL4(long m, long d, long n); /* MOV.L @(disp,Rm);,Rn */
  void DSMMOVA(int d);               /* MOVA @(disp,PC);,R0 */
  void DSMMOVCAL(int n);      //MOVCA.L  R0,@Rn
  void DSMMOVT(long n);               /* MOVT Rn */
  void DSMMULL(long m, long n); /* MUL.L Rm,Rn */
  void DSMMULS(long m, long n); /* MULS Rm,Rn */
  void DSMMULU(long m, long n); /* MULU Rm,Rn  */
  void DSMNEG(long m, long n); /* NEG Rm,Rn */
  void DSMNEGC(long m, long n); /* NEGC Rm,Rn */
  void DSMNOP(); /* NOP */
  void DSMNOT(long m, long n); /* NOT Rm,Rn */
  void DSMOCBI(int n);        /* OCBI @Rn */
  void DSMOCBP(int n);        /* OCBP @Rn */
  void DSMOCBWB(int n);       /* OCBWB @Rn */
  void DSMOR(long m, long n); /* OR Rm,Rn */
  void DSMORI(long i);  /* OR #imm,R0 */
  void DSMORM(long i); /* OR.B #imm,@(R0,GBR); */
  void DSMPREF(int n); /* PREF */
  void DSMROTCL(long n); /* ROTCL Rn */
  void DSMROTCR(long n); /* ROTCR Rn */
  void DSMROTL(long n);  /* ROTL Rn */
  void DSMROTR(long n); /* ROTR Rn */
  void DSMRTE(); /* RTE */
  void DSMRTS(); /* RTS */
  void DSMSETS(); /* SETS */
  void DSMSETT();  /* SETT */
  void DSMSHAD(int m,int n); /*SHAD Rm,Rn */
  void DSMSHAL(long n); /* SHAL Rn (Same as SHLL); */
  void DSMSHAR(long n); /* SHAR Rn */
  void DSMSHLD(int m,int n);/*SHLD Rm,Rn */
  void DSMSHLL(long n); /* SHLL Rn  (Same as SHAL); */
  void DSMSHLL2(long n); /* SHLL2 Rn */
  void DSMSHLL8(long n); /* SHLL8 Rn */
  void DSMSHLL16(long n); /* SHLL16 Rn */
  void DSMSHLR(long n);   /* SHLR Rn */
  void DSMSHLR2(long n);          /* SHLR2 Rn */
  void DSMSHLR8(long n);          /* SHLR8 Rn */
  void DSMSHLR16(long n);         /* SHLR16 Rn */
  void DSMSLEEP();    /* SLEEP */
  void DSMSTCSR(int n);      /* STC SR,Rn : Privileged */
  void DSMSTCGBR(int n);     /* STC GBR,Rn */
  void DSMSTCVBR(int n);     /* STC VBR,Rn : Privileged */
  void DSMSTCSSR(int n);     /* STC SSR,Rn : Privileged */
  void DSMSTCSPC(int n);     /* STC SPC,Rn : Privileged */
  void DSMSTCSGR(int n);     /* STC SGR,Rn : Privileged */
  void DSMSTCDBR(int n);     /* STC DBR,Rn : Privileged */
  void DSMSTCRm_BANK(int n, int m);     /* STC Rm_BANK,Rn : Privileged ; m=07 */
  void DSMSTCMSR(int n);     /* STC.L SR,@-Rn : Privileged */
  void DSMSTCMGBR(int n);     /* STC.L GBR,@Rn */
  void DSMSTCMVBR(int n);     /* STC.L VBR,@-Rn : Privileged */
  void DSMSTCMSSR(int n);     /* STC.L SSR,@-Rn : Privileged */
  void DSMSTCMSPC(int n);     /* STC.L SPC,@-Rn : Privileged */
  void DSMSTCMSGR(int n);     /* STC.L SGR,@-Rn : Privileged */
  void DSMSTCMDBR(int n);     /* STC.L DBR,@-Rn : Privileged */
  void DSMSTCMRm_BANK(int n, int m);     /* STC.L Rm_BANK,@-Rn : Privileged ; m=07 */
  void DSMSTSMACH(int n);     /* STS MACH,Rn */
  void DSMSTSMACL(int n);     /* STS MACL,Rn */
  void DSMSTSPR(int n);     /* STS PR,Rn */
  void DSMSTSMMACH(int n);     /* STS.L MACH,@-Rn */
  void DSMSTSMMACL(int n);     /* STS.L MACL,@-Rn */
  void DSMSTSMPR(int n);     /* STS.L PR,@-Rn */
  void DSMSTS(int n);           /* STS FPUL,Rn  */
  void DSMSTS_SAVE(int n);     /* STS.L FPUL,@-Rn  */
  void DSMSTS_FPSCR(int n);          /* STS FPSCR,Rn  */
  void DSMSTS_RESTORE(int n);  /* STS.L FPSCR,@-Rn  */
  void DSMSUB(long m, long n); /* SUB Rm,Rn */
  void DSMSUBC(long m, long n); /* SUBC Rm,Rn */
  void DSMSUBV(long m, long n); /* SUBV Rm,Rn */
  void DSMSWAPB(long m, long n);      /* SWAP.B Rm,Rn */
  void DSMSWAPW(long m, long n);      /* SWAP.W Rm,Rn */
  void DSMTAS(int n); /* TAS.B @Rn */
  void DSMTRAPA(int i); /* TRAPA #imm */
  void DSMTST(long m, long n); /* TST Rm,Rn */
  void DSMTSTI(long i); /* TST #imm,R0 */
  void DSMTSTM(long i); /* TST.B #imm,@(R0,GBR); */
  void DSMXOR(long m, long n); /* XOR Rm,Rn */
  void DSMXORI(long i);  /* XOR #imm,R0 */
  void DSMXORM(long i);  /* XOR.B #imm,@(R0,GBR); */
  void DSMXTRCT(long m, long n);     /* XTRCT Rm,Rn */
  void DSMFABS(int n);
  void DSMFLADD(int m,int n);
  void DSMFCMP_EQ(int m,int n); /* FCMP/EQ  FRm,FRn */
  void DSMFCMP_GT(int m,int n); /* FCMP/GT  FRm,FRn */
  void DSMFCNVDS(int m);
  void DSMFCNVSD(int n);
  void DSMFDIV(int m,int n);     /* FDIV FRm,FRn */
  void DSMFIPR(int m,int n);   /* FIPR FVm,FVn */
  void DSMFLDI0(int n);
  void DSMFLDI1(int n);
  void DSMFLDS(int m);
  void DSMFLOAT(int n);
  void DSMFMAC(int m,int n);
  void DSMFMUL(int m,int n);
  void DSMFNEG (int n);
  void DSMFSCHG();   /* FSCHG */
  void DSMFRCHG();   /* FRCHG */
  void DSMFSQRT(int n);
  void DSMFSTS(int n);
  void DSMFLSUB (int m,int n);
  void DSMFTRC(int m);
  void DSMFTRV (int n);     /* FTRV FVn */
  void DSMFMOV(int m,int n);
  void DSMFMOV_STORE(int m,int n);
  void DSMFMOV_LOAD(int m,int n);
  void DSMFMOV_RESTORE(int m,int n);
  void DSMFMOV_SAVE(int m,int n);
  void DSMFMOV_INDEX_LOAD(int m,int n);
  void DSMFMOV_INDEX_STORE(int m,int n);

//GTKGUI
extern unsigned long CODESIZE;
extern char *ROMNAME;
extern GtkWidget *pvrtext;
extern int frames;

#ifdef SH4DSM
 char *makereginfo(void);
 extern char dsmmsg[80];
 #define UNKNOWN_DSMINFO sprintf(dsmmsg,"?");
#endif

void mapall(void);



#define OPCALL(CPU_OP) \
void (*exeop)(int,int,int)=(void (*)(int,int,int))tab_op[CPU_OP].fonc;\
exeop(tab_op[CPU_OP].par1,tab_op[CPU_OP].par2,tab_op[CPU_OP].par3);


/*
"#ifdef _linux\
...
#elif WINDOWS\
 void (*exeop)(int,int,int)=tab_op[CPU_OP].fonc;\
 exeop(tab_op[CPU_OP].par1,tab_op[CPU_OP].par2,tab_op[CPU_OP].par3);\
#endif"
*/
