/************
Virtu@L Dream, DreamCast emulator for PC
Copyright (C) 2003 eBrain  H.H.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
************/

#include "StdAfx.h"
#include "extern.h"
//Descrambling routines are borrowed from a descramble.c
#define MAXCHUNK (2048*1024)

unsigned int seed;

void my_srand(unsigned int n)
{
  seed = n & 0xffff;
}

unsigned int my_rand()
{
  seed = (seed * 2109 + 9273) & 0x7fff;
  return (seed + 0xc000) & 0xffff;
}

void load(FILE *fh, unsigned char *ptr, unsigned long sz)
{
  if(fread(ptr, 1, sz, fh) != sz)
    {
      fprintf(stderr, "Read error!\n");
      exit(1);
    }
}

void load_chunk(FILE *fh, unsigned char *ptr, unsigned long sz)
{
  int idx[MAXCHUNK/32];
  long i;
  long x,tmp;

  /* Convert chunk size to number of slices */
  sz /= 32;

  /* Initialize index table with unity,
     so that each slice gets loaded exactly once */
  for(i = 0; i < sz; i++)
    idx[i] = i;

  for(i = sz-1; i >= 0; --i)
    {
      /* Select a replacement index */
      x = (my_rand() * i) >> 16;

      /* Swap */
      tmp = idx[i];
      idx[i] = idx[x];
      idx[x] = tmp;

      /* Load resulting slice */
      load(fh, ptr+32*idx[i], 32);
    }
}

void load_file(FILE *fh,unsigned char *ptr, unsigned long filesz)
{
  unsigned long chunksz;

  my_srand(filesz);

  /* Descramble 2 meg blocks for as long as possible, then
     gradually reduce the window down to 32 bytes (1 slice) */
  for(chunksz = MAXCHUNK; chunksz >= 32; chunksz >>= 1)
    while(filesz >= chunksz)
      {
        load_chunk(fh, ptr, chunksz);
        filesz -= chunksz;
        ptr += chunksz;
      }
  /* Load final incomplete slice */
  if(filesz)
    load(fh, ptr, filesz);
}

unsigned char *read_file(char *filename)
{
  unsigned long sz;
  unsigned char *retbuf;
  FILE *fh = fopen(filename, "rb");
  if(fh == NULL)
    {
      fprintf(stderr, "Can't open \"%s\".\n", filename);
      exit(1);
    }
  if(fseek(fh, 0, SEEK_END)<0)
    {
      fprintf(stderr, "Seek error.\n");
      exit(1);
    }
  sz = ftell(fh);
  retbuf = (unsigned char *) malloc(sz);
  if(retbuf == NULL )
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }
  if(fseek(fh, 0, SEEK_SET)<0)
    {
      fprintf(stderr, "Seek error.\n");
      exit(1);
    }
  printf("mem alloc, sz:%d\n",sz);
  load_file(fh, retbuf, sz);
  printf("got desc\n");
  fclose(fh);
  return retbuf;
}

unsigned char *descramble(char *src)
{
  return (read_file(src));
}


void max2min(char *filename_case)
{
 int i=0;
 while (filename_case[i]!=0)
 {
  if (('A'<=filename_case[i]) && (filename_case[i]<='Z'))
  {
   filename_case[i]+=32;
  }
  i++;
 }
}


void chop(char *str)
{
 while (*str)
 {
  if ((*str==' ') && (*(str+sizeof(char))==' '))
  {
   *str=0;
   break;
  }
  else
  {
   str++;
  }
 }
}

void loadip_1stread(char *ipbinfile)
{
typedef struct
{
 long offset;
 int len;
} ROMINFO ;

//META INFO
 ROMINFO HardwareID;
 ROMINFO MakerID;
 ROMINFO DevInfo;
 ROMINFO AreaSymbol;
 ROMINFO Periph;
 ROMINFO ProductNum;
 ROMINFO ProductVer;
 ROMINFO ReleaseDate;
 ROMINFO BootFile;
 ROMINFO CompanyName;
 ROMINFO SoftName;
//
 ROMINFO MetaInfo;
 ROMINFO TableOfContent;
 ROMINFO SEGALicenceScreenCode;
 ROMINFO AreaProtectionSymbole;
 ROMINFO Bootstrap1;
 ROMINFO Bootstrap2;

 FILE *h_ip_bin;
 FILE *h_bin;
 BYTE *mem_ip_bin;
 unsigned long fsize;
 int total_loaded;
 char binread1stfilename[0x10];
 unsigned char *desc;
 char desrom;


 h_ip_bin=fopen(ipbinfile,"r");
 mem_ip_bin=DREAMCAST_PRAM+IPOFFSET;
 fread(mem_ip_bin,1,0x8000,h_ip_bin);
 fclose(h_ip_bin);

 HardwareID.offset=0;
 HardwareID.len=0x10;
 MakerID.offset=0x10;
 MakerID.len=0x10;
 DevInfo.offset=0x20;
 DevInfo.len=0x10;
 AreaSymbol.offset=0x30;
 AreaSymbol.len=8;
 Periph.offset=0x38;
 Periph.len=8;
 ProductNum.offset=0x40;
 ProductNum.len=10;
 ProductVer.offset=0x4a;
 ProductVer.len=6;
 ReleaseDate.offset=0x50;
 ReleaseDate.len=0x10;
 BootFile.offset=0x60;
 BootFile.len=0x10;

 sscanf((const char*)(BootFile.offset+mem_ip_bin),"%s",binread1stfilename);
 printf("%s is also required\n",binread1stfilename);

 CompanyName.offset=0x70;
 CompanyName.len=0x10;
 SoftName.offset=0x80;
 SoftName.len=0x10;

 if (ROMNAME) free(ROMNAME);
 ROMNAME=(char *)malloc(strlen((const char *)((int)SoftName.offset+(int)mem_ip_bin))+1);
 strcpy(ROMNAME,(const char *)((int)SoftName.offset+(int)mem_ip_bin));
 chop(ROMNAME);

 MetaInfo.offset=0;
 MetaInfo.len=0x100;
 TableOfContent.offset=0x100;
 TableOfContent.len=0x200;
 SEGALicenceScreenCode.offset=0x300;
 SEGALicenceScreenCode.len=0x3400;
 AreaProtectionSymbole.offset=0x3700;
 AreaProtectionSymbole.len=0x100;
 Bootstrap1.offset=0x3800;
 Bootstrap1.len=0x2800;
 Bootstrap2.offset=0x6000;
 Bootstrap2.len=0x2000;

 PC=0x8c008300;
 GBR=0x8C000000;
 VBR=0x8C000000;

 h_bin=fopen(binread1stfilename,"r");
 desrom=0;
 //linux c strcasecmp
 #if defined _linux
  if (!(strcasecmp(binread1stfilename,"1st_read.bin"))) desrom=1;
 #elif defined WINDOWS
  if (!(stricmp(binread1stfilename,"1st_read.bin"))) desrom=1;
 #endif
 if (!h_bin)
 {
  max2min(binread1stfilename);
  printf("trying %s\n",binread1stfilename);
  h_bin=fopen(binread1stfilename,"r");
  if (!h_bin)
  {
   printf("%s open err\n",binread1stfilename);
   exit(-1);
  }
 }
 fseek(h_bin,0,SEEK_END);
 fsize=ftell(h_bin);
 CODESIZE=fsize;
 mem_ip_bin=DREAMCAST_PRAM+BOOTOFFSET;
 printf("%d bytes for 1st_read rom ",CODESIZE);

 if (desrom)
 {
  fclose(h_bin);
  printf("descrambling...\n");
  desc=descramble(binread1stfilename);
  printf("copy to memory...\n");
  memcpy(mem_ip_bin,desc,fsize);
  free(desc);
 }
 else
 {
  /* sans desc */
  printf("loading without descrambling...\n");
  fseek(h_bin,0,SEEK_SET);
  fread(mem_ip_bin,1,fsize,h_bin);
  fclose(h_bin);
 }

 printf("ok\n");
}
